@@ -1,12 +1,12 @@
 /*=========================================================================
  *
- *  Copyright Insight Software Consortium
+ *  Copyright NumFOCUS
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
  *  You may obtain a copy of the License at
  *
- *         http://www.apache.org/licenses/LICENSE-2.0.txt
+ *         https://www.apache.org/licenses/LICENSE-2.0.txt
  *
  *  Unless required by applicable law or agreed to in writing, software
  *  distributed under the License is distributed on an "AS IS" BASIS,
@@ -87,30 +87,29 @@ namespace itk
  * coordinates of the point are set via SetPointData. For example, to compute a
  * spline through the (ordered) 2D points (5,6) and (7,8), you should use:
  *
- * \code
- * typedef itk::Vector< float, 2 > DataType;
- * PointSetType::PointType param0;
- * param0[0] = 0.0;
- * DataType p0;
- * p0[0] =  10.0; p0[1]= 10.0;
- * pointSet->SetPoint(0, param0);
- * pointSet->SetPointData( 0, p0 );
- *
- * PointSetType::PointType param1;
- * param1[0] = 1.0;
- * DataType p1;
- * p1[0] =  80.0; p1[1]= 50.0;
- * pointSet->SetPoint(1, param1);
- * pointSet->SetPointData( 1, p1 );
- * \endcode
+   \code
+   using DataType = itk::Vector< float, 2 >;
+   PointSetType::PointType param0;
+   param0[0] = 0.0;
+   DataType p0;
+   p0[0] =  10.0; p0[1]= 10.0;
+   pointSet->SetPoint(0, param0);
+   pointSet->SetPointData( 0, p0 );
+
+   PointSetType::PointType param1;
+   param1[0] = 1.0;
+   DataType p1;
+   p1[0] =  80.0; p1[1]= 50.0;
+   pointSet->SetPoint(1, param1);
+   pointSet->SetPointData( 1, p1 );
+   \endcode
  *
  * \author Nicholas J. Tustison
  *
  * This code was contributed in the Insight Journal paper:
  * "N-D C^k B-Spline Scattered Data Approximation"
  * by Nicholas J. Tustison, James C. Gee
- * https://hdl.handle.net/1926/140
- * http://www.insight-journal.org/browse/publication/57
+ * https://www.insight-journal.org/browse/publication/57
  *
  *
  * \par REFERENCE
@@ -123,115 +122,121 @@ namespace itk
  * with Confidence Values", Proceedings of the MIAR conference, August 2006.
  *
  * \ingroup ITKImageGrid
+ *
+ * \sphinx
+ * \sphinxexample{Filtering/ImageGrid/FitSplineIntoPointSet,Fit Spline Into Point Set}
+ * \endsphinx
  */
 
-template< typename TInputPointSet, typename TOutputImage >
-class ITK_TEMPLATE_EXPORT BSplineScatteredDataPointSetToImageFilter:
-  public PointSetToImageFilter< TInputPointSet, TOutputImage >
+template <typename TInputPointSet, typename TOutputImage>
+class ITK_TEMPLATE_EXPORT BSplineScatteredDataPointSetToImageFilter
+  : public PointSetToImageFilter<TInputPointSet, TOutputImage>
 {
 public:
-  /** Standard class typedefs. */
-  typedef BSplineScatteredDataPointSetToImageFilter             Self;
-  typedef PointSetToImageFilter<TInputPointSet, TOutputImage>   Superclass;
-  typedef SmartPointer<Self>                                    Pointer;
-  typedef SmartPointer<const Self>                              ConstPointer;
+  ITK_DISALLOW_COPY_AND_MOVE(BSplineScatteredDataPointSetToImageFilter);
+
+  /** Standard class type aliases. */
+  using Self = BSplineScatteredDataPointSetToImageFilter;
+  using Superclass = PointSetToImageFilter<TInputPointSet, TOutputImage>;
+  using Pointer = SmartPointer<Self>;
+  using ConstPointer = SmartPointer<const Self>;
 
   /** Method for creation through the object factory. */
-  itkNewMacro( Self );
+  itkNewMacro(Self);
 
-  /** Run-time type information (and related methods). */
-  itkTypeMacro( BSplineScatteredDataPointSetToImageFilter,
-    PointSetToImageFilter );
+  /** \see LightObject::GetNameOfClass() */
+  itkOverrideGetNameOfClassMacro(BSplineScatteredDataPointSetToImageFilter);
 
   /** Extract dimension from the output image. */
-  itkStaticConstMacro( ImageDimension, unsigned int,
-    TOutputImage::ImageDimension );
+  static constexpr unsigned int ImageDimension = TOutputImage::ImageDimension;
 
-  typedef TOutputImage                              ImageType;
-  typedef TInputPointSet                            PointSetType;
+  using ImageType = TOutputImage;
+  using PointSetType = TInputPointSet;
 
-  /** Image typedef support. */
-  typedef typename ImageType::PixelType             PixelType;
-  typedef typename ImageType::RegionType            RegionType;
-  typedef typename ImageType::SizeType              SizeType;
-  typedef typename ImageType::IndexType             IndexType;
+  /** Image type alias support */
+  using PixelType = typename ImageType::PixelType;
+  using RegionType = typename ImageType::RegionType;
+  using OutputImageRegionType = RegionType;
+  using SizeType = typename ImageType::SizeType;
+  using IndexType = typename ImageType::IndexType;
 
-  /** PointSet typedef support. */
-  typedef typename PointSetType::PointType          PointType;
-  typedef typename PointSetType::Pointer            PointSetPointer;
-  typedef typename PointSetType::PixelType          PointDataType;
-  typedef typename PointSetType::PointDataContainer PointDataContainerType;
+  /** PointSet type alias support */
+  using PointType = typename PointSetType::PointType;
+  using PointSetPointer = typename PointSetType::Pointer;
+  using PointDataType = typename PointSetType::PixelType;
+  using PointDataContainerType = typename PointSetType::PointDataContainer;
+  using PointDataContainerPointer = typename PointDataContainerType::Pointer;
 
-  /** Other typedefs. */
-  typedef float                                     RealType;
-  typedef VectorContainer<unsigned, RealType>       WeightsContainerType;
+  /** Other type alias. */
+  using RealType = float;
+  using WeightsContainerType = VectorContainer<unsigned int, RealType>;
 
   /** Image types. */
-  typedef Image<PointDataType,
-    itkGetStaticConstMacro( ImageDimension )>       PointDataImageType;
-  typedef Image<RealType,
-    itkGetStaticConstMacro( ImageDimension )>       RealImageType;
-  typedef typename RealImageType::Pointer           RealImagePointer;
-  typedef typename PointDataImageType::Pointer      PointDataImagePointer;
-  typedef FixedArray<unsigned,
-    itkGetStaticConstMacro( ImageDimension )>       ArrayType;
-  typedef FixedArray<RealType,
-    itkGetStaticConstMacro( ImageDimension )>       RealArrayType;
+  using PointDataImageType = Image<PointDataType, Self::ImageDimension>;
+  using RealImageType = Image<RealType, Self::ImageDimension>;
+  using RealImagePointer = typename RealImageType::Pointer;
+  using PointDataImagePointer = typename PointDataImageType::Pointer;
+  using ArrayType = FixedArray<unsigned int, Self::ImageDimension>;
+  using RealArrayType = FixedArray<RealType, Self::ImageDimension>;
 
   /** Interpolation kernel type (default spline order = 3). */
-  typedef CoxDeBoorBSplineKernelFunction<3>         KernelType;
-  typedef BSplineKernelFunction<0>                  KernelOrder0Type;
-  typedef BSplineKernelFunction<1>                  KernelOrder1Type;
-  typedef BSplineKernelFunction<2>                  KernelOrder2Type;
-  typedef BSplineKernelFunction<3>                  KernelOrder3Type;
+  using KernelType = CoxDeBoorBSplineKernelFunction<3>;
+  using KernelOrder0Type = BSplineKernelFunction<0>;
+  using KernelOrder1Type = BSplineKernelFunction<1>;
+  using KernelOrder2Type = BSplineKernelFunction<2>;
+  using KernelOrder3Type = BSplineKernelFunction<3>;
 
 
   /** Set the spline order assuming it is the same in all parametric dimensions.
    * The spline order determines the continuity between B-spline elements and
    * the degree of polynomial used to construct the B-spline elements. Default
    * = 3. */
-  void SetSplineOrder( unsigned int );
+  void
+  SetSplineOrder(unsigned int);
 
   /** Set the spline order for each parametric dimension separately. The spline
    * order determines the continuity between B-spline elements and the degree of
    * polynomial used to construct the B-spline elements. Default = 3. */
-  void SetSplineOrder( const ArrayType & );
+  void
+  SetSplineOrder(const ArrayType &);
 
   /** Get the spline order for all parametric dimensions. The spline order
    * determines the continuity between B-spline elements and the degree of
    * polynomial used to construct the B-spline elements. Default = 3. */
-  itkGetConstReferenceMacro( SplineOrder, ArrayType );
+  itkGetConstReferenceMacro(SplineOrder, ArrayType);
 
   /** Set/Get the number of control points for each parametric dimension at the
    * initial fitting level. The B-spline mesh size is equal to the number
    * of control points minus the spline order. Default = 4 in each dimension.
    */
-  itkSetMacro( NumberOfControlPoints, ArrayType );
-  itkGetConstReferenceMacro( NumberOfControlPoints, ArrayType );
+  itkSetMacro(NumberOfControlPoints, ArrayType);
+  itkGetConstReferenceMacro(NumberOfControlPoints, ArrayType);
 
   /** Get the number of current control points for each parametric dimension at
    * the current fitting level. The B-spline mesh size is equal to the number
    * of control points minus the spline order. Default = 4 in each dimension.
    */
-  itkGetConstReferenceMacro( CurrentNumberOfControlPoints, ArrayType );
+  itkGetConstReferenceMacro(CurrentNumberOfControlPoints, ArrayType);
 
   /** Set the number of fitting levels assuming the number of fitting levels is
    * the same for each parametric dimension. Starting with the mesh size
    * implied by setting the number of control points, the mesh size is doubled
    * at each fitting level. Default = 1 in all parametric dimensions. */
-  void SetNumberOfLevels( unsigned int );
+  void
+  SetNumberOfLevels(unsigned int);
 
   /** Set the number of fitting levels in each parametric dimension separately.
    * Starting with the mesh size implied by setting the number of control
    * points, the mesh size is doubled at each fitting level. Default = 1 in all
    * parametric dimensions. */
-  void SetNumberOfLevels( const ArrayType & );
+  void
+  SetNumberOfLevels(const ArrayType &);
 
   /** Get the number of fitting levels for all parametric dimensions. Starting
    * with the mesh size implied by setting the number of control points, the
    * mesh size is doubled at each fitting level. Default = 1 in all parametric
    * dimensions. */
-  itkGetConstReferenceMacro( NumberOfLevels, ArrayType );
+  itkGetConstReferenceMacro(NumberOfLevels, ArrayType);
 
   /** Set/Get the epsilon used for B-splines. The B-spline parametric domain in
    * 1-D is defined on the half-closed interval [a,b). Extension to n-D is
@@ -239,8 +244,8 @@ public:
    * the image domain to be co-extensive with the parametric domain. We use
    * the B-spline epsilon to push the edge of the image boundary inside the
    * B-spline parametric domain. */
-  itkSetMacro( BSplineEpsilon, RealType );
-  itkGetConstMacro( BSplineEpsilon, RealType );
+  itkSetMacro(BSplineEpsilon, RealType);
+  itkGetConstMacro(BSplineEpsilon, RealType);
 
   /** Set/Get the array to define the periodicity of the dimensions in the
    * parametric space is to be.
@@ -252,120 +257,138 @@ public:
    * cylindrical topology, the array type will have two components, and you
    * should set to "1" the component that goes around the cylinder, and set to
    * "0" the component that goes from the top of the cylinder to the bottom.
-   * This will indicate the periodity of that parameter to the filter.
+   * This will indicate the periodicity of that parameter to the filter.
    * Internally, in order to make periodic the domain of the parameter, the
    * filter will reuse some of the points at the beginning of the domain as if
    * they were also located at the end of the domain. The number of points to
    * be reused will depend on the spline order. As a user, you don't need to
    * replicate the points, the filter will do this for you. */
-  itkSetMacro( CloseDimension, ArrayType );
-  itkGetConstReferenceMacro( CloseDimension, ArrayType );
+  itkSetMacro(CloseDimension, ArrayType);
+  itkGetConstReferenceMacro(CloseDimension, ArrayType);
 
   /** A weighted fitting is possible where each input point is assigned a
    * relative weighting. */
-  void SetPointWeights( WeightsContainerType *weights );
+  void
+  SetPointWeights(WeightsContainerType * weights);
 
   /** Set/Get whether or not the sampled output B-spline object is constructed.
    * The result of the fitting process is an n-D grid of control points which
    * describe the continuous B-spline object. */
-  itkSetMacro( GenerateOutputImage, bool );
-  itkGetConstReferenceMacro( GenerateOutputImage, bool );
-  itkBooleanMacro( GenerateOutputImage );
+  itkSetMacro(GenerateOutputImage, bool);
+  itkGetConstReferenceMacro(GenerateOutputImage, bool);
+  itkBooleanMacro(GenerateOutputImage);
 
   /** Get the control point lattice produced by the fitting process. */
-  PointDataImagePointer GetPhiLattice()
-    {
-    return static_cast<PointDataImageType *>( this->ProcessObject::GetOutput( 1 ) );
-    }
+  PointDataImagePointer
+  GetPhiLattice()
+  {
+    return static_cast<PointDataImageType *>(this->ProcessObject::GetOutput(1));
+  }
 
 protected:
   BSplineScatteredDataPointSetToImageFilter();
-  virtual ~BSplineScatteredDataPointSetToImageFilter() ITK_OVERRIDE;
+  ~BSplineScatteredDataPointSetToImageFilter() override = default;
 
-  void PrintSelf(std::ostream & os, Indent indent) const ITK_OVERRIDE;
+  void
+  PrintSelf(std::ostream & os, Indent indent) const override;
 
-  void ThreadedGenerateData( const RegionType &, ThreadIdType ) ITK_OVERRIDE;
+  void
+  ThreadedGenerateData(const RegionType &, ThreadIdType) override;
 
-  void BeforeThreadedGenerateData() ITK_OVERRIDE;
+  void
+  DynamicThreadedGenerateData(const RegionType &) override
+  {
+    itkExceptionMacro("This class requires threadId so it must use classic multi-threading model");
+  }
 
-  void AfterThreadedGenerateData() ITK_OVERRIDE;
+  void
+  BeforeThreadedGenerateData() override;
 
-  unsigned int SplitRequestedRegion( unsigned int, unsigned int, RegionType & ) ITK_OVERRIDE;
+  void
+  AfterThreadedGenerateData() override;
 
-  void GenerateData() ITK_OVERRIDE;
+  unsigned int
+  SplitRequestedRegion(unsigned int, unsigned int, RegionType &) override;
+
+  void
+  GenerateData() override;
 
 private:
-
-  ITK_DISALLOW_COPY_AND_ASSIGN(BSplineScatteredDataPointSetToImageFilter);
-
   /** Function used to propagate the fitting solution at one fitting level
    * to the next level with the mesh resolution doubled. */
-  void RefineControlPointLattice();
-
-  /** Determine the residuals after fitting to one level. */
-  void UpdatePointSet();
+  void
+  RefineControlPointLattice();
 
   /** This function is not used as it requires an evaluation of all
    * (SplineOrder+1)^ImageDimensions B-spline weights for each evaluation. */
-  void GenerateOutputImage();
+  void
+  GenerateOutputImage();
 
   /** Function used to generate the sampled B-spline object quickly. */
-  void ThreadedGenerateDataForFitting( const RegionType &, ThreadIdType );
+  void
+  ThreadedGenerateDataForFitting(const RegionType &, ThreadIdType);
 
   /** Function used to generate the sampled B-spline object quickly. */
-  void ThreadedGenerateDataForReconstruction( const RegionType &, ThreadIdType );
+  void
+  ThreadedGenerateDataForReconstruction(const RegionType &, ThreadIdType);
+
+  /** Update the residuals for multi-level fitting. */
+  void
+  ThreadedGenerateDataForUpdatingResidualValues(const RegionType &, ThreadIdType);
 
   /** Sub-function used by GenerateOutputImageFast() to generate the sampled
    * B-spline object quickly. */
-  void CollapsePhiLattice( PointDataImageType *, PointDataImageType *,
-    const RealType, const unsigned int );
+  void
+  CollapsePhiLattice(PointDataImageType *, PointDataImageType *, const RealType, const unsigned int);
 
   /** Set the grid parametric domain parameters such as the origin, size,
    * spacing, and direction. */
-  void SetPhiLatticeParametricDomainParameters();
+  void
+  SetPhiLatticeParametricDomainParameters();
 
   /** Convert number to index given a size of image. Used to index
    * the local control point neighborhoods. */
-  IndexType NumberToIndex( const unsigned int, const SizeType );
+  IndexType
+  NumberToIndex(const unsigned int, const SizeType);
 
-  bool                                         m_DoMultilevel;
-  bool                                         m_GenerateOutputImage;
-  bool                                         m_UsePointWeights;
-  unsigned int                                 m_MaximumNumberOfLevels;
-  unsigned int                                 m_CurrentLevel;
-  ArrayType                                    m_NumberOfControlPoints;
-  ArrayType                                    m_CurrentNumberOfControlPoints;
-  ArrayType                                    m_CloseDimension;
-  ArrayType                                    m_SplineOrder;
-  ArrayType                                    m_NumberOfLevels;
+  bool         m_DoMultilevel{ false };
+  bool         m_GenerateOutputImage{ true };
+  bool         m_UsePointWeights{ false };
+  unsigned int m_MaximumNumberOfLevels{ 1 };
+  unsigned int m_CurrentLevel{ 0 };
+  ArrayType    m_NumberOfControlPoints{};
+  ArrayType    m_CurrentNumberOfControlPoints{};
+  ArrayType    m_CloseDimension{};
+  ArrayType    m_SplineOrder{};
+  ArrayType    m_NumberOfLevels{};
 
-  typename WeightsContainerType::Pointer       m_PointWeights;
+  typename WeightsContainerType::Pointer m_PointWeights{};
 
-  typename PointDataImageType::Pointer         m_PhiLattice;
-  typename PointDataImageType::Pointer         m_PsiLattice;
+  typename PointDataImageType::Pointer m_PhiLattice{};
+  typename PointDataImageType::Pointer m_PsiLattice{};
 
-  vnl_matrix<RealType>     m_RefinedLatticeCoefficients[ImageDimension];
+  vnl_matrix<RealType> m_RefinedLatticeCoefficients[ImageDimension]{};
 
-  typename PointDataContainerType::Pointer     m_InputPointData;
-  typename PointDataContainerType::Pointer     m_OutputPointData;
+  PointDataContainerPointer m_ResidualPointSetValues{};
 
-  typename KernelType::Pointer                 m_Kernel[ImageDimension];
+  typename KernelType::Pointer m_Kernel[ImageDimension]{};
 
-  typename KernelOrder0Type::Pointer           m_KernelOrder0;
-  typename KernelOrder1Type::Pointer           m_KernelOrder1;
-  typename KernelOrder2Type::Pointer           m_KernelOrder2;
-  typename KernelOrder3Type::Pointer           m_KernelOrder3;
+  typename KernelOrder0Type::Pointer m_KernelOrder0{};
+  typename KernelOrder1Type::Pointer m_KernelOrder1{};
+  typename KernelOrder2Type::Pointer m_KernelOrder2{};
+  typename KernelOrder3Type::Pointer m_KernelOrder3{};
 
-  std::vector<RealImagePointer>                m_OmegaLatticePerThread;
-  std::vector<PointDataImagePointer>           m_DeltaLatticePerThread;
+  std::vector<RealImagePointer>      m_OmegaLatticePerThread{};
+  std::vector<PointDataImagePointer> m_DeltaLatticePerThread{};
 
-  RealType                                     m_BSplineEpsilon;
-  bool                                         m_IsFittingComplete;
+  RealType m_BSplineEpsilon{ static_cast<RealType>(1e-3) };
+  bool     m_IsFittingComplete{ false };
+  bool     m_DoUpdateResidualValues{ false };
 };
 } // end namespace itk
 
 #ifndef ITK_MANUAL_INSTANTIATION
-#include "itkBSplineScatteredDataPointSetToImageFilter.hxx"
+#  include "itkBSplineScatteredDataPointSetToImageFilter.hxx"
 #endif
 
 #endif
